import unittest
from nose.tools import *

from bs4 import BeautifulSoup

from robobrowser import helpers


class TestEnsureSoup(unittest.TestCase):
    
    def setUp(self):
        self.html = '<form></form>'
        self.tag = BeautifulSoup(self.html).find()
        self.htmls = [
            '<form></form>',
            '<input />',
        ]
        self.tags = [
            BeautifulSoup(html).find()
            for html in self.htmls
        ]

    def test_handle_string(self):
        ensured = helpers.ensure_soup(self.html)
        assert_equal(ensured, self.tag)
    
    def test_handle_soup(self):
        ensured = helpers.ensure_soup(BeautifulSoup(self.html))
        assert_equal(ensured, self.tag)

    def test_handle_tag(self):
        ensured = helpers.ensure_soup(BeautifulSoup(self.html).find())
        assert_equal(ensured, self.tag)

    def test_handle_string_list(self):
        ensured = helpers.ensure_soup(self.htmls)
        assert_equal(ensured, self.tags)

    def test_handle_soup_list(self):
        ensured = helpers.ensure_soup([
            BeautifulSoup(html)
            for html in self.htmls
        ])
        assert_equal(ensured, self.tags)

    def test_handle_tag_list(self):
        ensured = helpers.ensure_soup([
            BeautifulSoup(html).find()
            for html in self.htmls
        ])
        assert_equal(ensured, self.tags)




from robobrowser import responses

from tests import utils

mock_links = utils.mock_responses(
    [
        utils.ArgCatcher(
            responses.GET, 'http://robobrowser.com/links/',
            body=b'''
                <a href="/link1/">sheer heart attack</a>
                <a href="/link2/" class="song">night at the opera</a>
                <a class="nohref">no href</a>
            '''
        ),
        utils.ArgCatcher(responses.GET, 'http://robobrowser.com/link1/'),
        utils.ArgCatcher(responses.GET, 'http://robobrowser.com/link2/'),
    ]
)

mock_forms = utils.mock_responses(
    [
        utils.ArgCatcher(
            responses.GET, 'http://robobrowser.com/get_form/',
            body=b'''
                <form id="bass" method="get" action="/get_form/">'
                    <input name="deacon" value="john" />
                </form>
                <form id="drums" method="post" action="/get_form/">'
                    <input name="deacon" value="john" />
                </form>
            '''
        ),
        utils.ArgCatcher(
            responses.GET, 'http://robobrowser.com/multi_submit_form/',
            body=b'''
                <form id="bass" method="get" action="/multi_submit_form/">'
                    <input name="deacon" value="john" />
                    <input type="submit" name="submit1" value="value1" />
                    <input type="submit" name="submit2" value="value2" />
                </form>
            '''
        ),
        utils.ArgCatcher(
            responses.GET, 'http://robobrowser.com/post_form/',
            body=b'''
                <form id="bass" method="post" action="/submit/">'
                    <input name="deacon" value="john" />
                </form>
                <form id="drums" method="post" action="/submit/">'
                    <input name="deacon" value="john" />
                </form>
            '''
        ),
        utils.ArgCatcher(
            responses.GET, 'http://robobrowser.com/noname/',
            body=b'''
                <form name="input" action="action" method="get">
                <input type="checkbox" name="vehicle" value="Bike">
                    I have a bike<br>
                <input type="checkbox" name="vehicle" value="Car">I have a car
                <br><br>
                <input type="submit" value="Submit">
                </form>
            '''
        ),
        utils.ArgCatcher(
            responses.POST, 'http://robobrowser.com/submit/',
        ),
    ]
)

mock_urls = utils.mock_responses(
    [
        utils.ArgCatcher(responses.GET, 'http://robobrowser.com/page1/'),
        utils.ArgCatcher(responses.GET, 'http://robobrowser.com/page2/'),
        utils.ArgCatcher(responses.GET, 'http://robobrowser.com/page3/'),
        utils.ArgCatcher(responses.GET, 'http://robobrowser.com/page4/'),
    ]
)



import functools

from robobrowser import responses
from robobrowser.compat import iteritems


class ArgCatcher(object):
    """Simple class for memorizing positional and keyword arguments. Used to
    capture responses for mock_responses.

    """
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


class KwargSetter(object):
    """Simple class for memorizing keyword arguments as instance attributes.
    Used to mock requests and responses for testing.

    """
    def __init__(self, **kwargs):
        for key, value in iteritems(kwargs):
            setattr(self, key, value)


def mock_responses(resps):
    """Decorator factory to make tests more DRY. Bundles responses.activate
    with a collection of response rules.

    :param list resps: List of response-formatted ArgCatcher arguments.

    """
    def wrapper(func):
        @responses.activate
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            for resp in resps:
                responses.add(*resp.args, **resp.kwargs)
            return func(*args, **kwargs)
        return wrapped
    return wrapper



import sys

PY2 = int(sys.version[0]) == 2
PY26 = PY2 and int(sys.version_info[1]) < 7

if PY26:
    from .ordereddict import OrderedDict
else:
    from collections import OrderedDict
OrderedDict = OrderedDict

if PY2:
    import urlparse
    urlparse = urlparse
    string_types = (str, unicode)
    unicode = unicode
    basestring = basestring
    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    iteritems = lambda d: d.iteritems()
    builtin_name = '__builtin__'
else:
    import urllib.parse
    urlparse = urllib.parse
    string_types = (str,)
    unicode = str
    basestring = (str, bytes)
    iterkeys = lambda d: iter(d.keys())
    itervalues = lambda d: iter(d.values())
    iteritems = lambda d: iter(d.items())
    builtin_name = 'builtins'



import unittest
from nose.tools import *

import datetime

from robobrowser.browser import RoboBrowser
from robobrowser.cache import RoboCache
from tests.utils import KwargSetter


class TestAdapter(unittest.TestCase):

    def test_cache_on(self):
        self.browser = RoboBrowser(cache=True)
        self.browser.open('http://httpbin.org/')
        resp1 = self.browser.state.response
        self.browser.open('http://httpbin.org/')
        resp2 = self.browser.state.response
        assert_true(resp1 is resp2)

    def test_cache_off(self):
        self.browser = RoboBrowser(cache=False)
        self.browser.open('http://httpbin.org/')
        resp1 = self.browser.state.response
        self.browser.open('http://httpbin.org/')
        resp2 = self.browser.state.response
        assert_true(resp1 is not resp2)


class TestCache(unittest.TestCase):

    def setUp(self):
        self.cache = RoboCache()

    def test_store(self):
        url = 'http://robobrowser.com/'
        response = KwargSetter(url=url, status_code=200)
        now = datetime.datetime.now()
        self.cache.store(response)
        assert_true(url in self.cache.data)
        assert_equal(self.cache.data[url]['response'], response)
        date_diff = self.cache.data[url]['date'] - now
        assert_true(date_diff < datetime.timedelta(seconds=0.1))

    def test_store_invalid_verb(self):
        url = 'http://robobrowser.com/'
        response = KwargSetter(url=url, status_code=400)
        self.cache.store(response)
        assert_false(url in self.cache.data)

    def test_retrieve_not_stored(self):
        request = KwargSetter(url='http://robobrowser.com/', method='GET')
        retrieved = self.cache.retrieve(request)
        assert_equal(retrieved, None)

    def test_retrieve_stored(self):
        request = KwargSetter(url='http://robobrowser.com/', method='GET')
        response = KwargSetter(url='http://robobrowser.com/', status_code=200)
        self.cache.store(response)
        retrieved = self.cache.retrieve(request)
        assert_equal(retrieved, response)

    def test_retrieve_invalid_code(self):
        request = KwargSetter(url='http://robobrowser.com/', method='GET')
        response = KwargSetter(url='http://robobrowser.com/', status_code=400)
        self.cache.store(response)
        retrieved = self.cache.retrieve(request)
        assert_equal(retrieved, None)

    def test_reduce_age(self):
        for idx in range(5):
            response = KwargSetter(url=idx, status_code=200)
            self.cache.store(response)
            # time.sleep(0.1)
        assert_equal(len(self.cache.data), 5)
        now = datetime.datetime.now()
        self.cache.max_age = now - self.cache.data[2]['date']
        self.cache._reduce_age(now)
        assert_equal(len(self.cache.data), 3)
        # Cast keys to list for 3.3 compatibility
        assert_equal(list(self.cache.data.keys()), [2, 3, 4])

    def test_reduce_count(self):
        for idx in range(5):
            response = KwargSetter(url=idx, status_code=200)
            self.cache.store(response)
        assert_equal(len(self.cache.data), 5)
        self.cache.max_count = 3
        self.cache._reduce_count()
        assert_equal(len(self.cache.data), 3)
        # Cast keys to list for 3.3 compatibility
        assert_equal(list(self.cache.data.keys()), [2, 3, 4])



import mock
import unittest
from nose.tools import *

import re
import requests

from robobrowser.browser import RoboBrowser
from robobrowser import exceptions

from tests.fixtures import mock_links, mock_urls, mock_forms


class TestHeaders(unittest.TestCase):

    @mock_links
    def test_headers(self):
        headers = {
            'X-Song': 'Innuendo',
            'X-Writer': 'Freddie',
        }
        browser = RoboBrowser(headers=headers)
        browser.open('http://robobrowser.com/links/')
        for key, value in headers.items():
            assert_equal(browser.session.headers[key], value)

    @mock_links
    def test_user_agent(self):
        browser = RoboBrowser(user_agent='freddie')
        browser.open('http://robobrowser.com/links/')
        assert_true('User-Agent' in browser.session.headers)
        assert_equal(
            browser.session.headers['User-Agent'], 'freddie'
        )

    def test_default_headers(self):
        browser = RoboBrowser()
        assert_equal(browser.session.headers, requests.Session().headers)


class TestLinks(unittest.TestCase):

    @mock_links
    def setUp(self):
        self.browser = RoboBrowser()
        self.browser.open('http://robobrowser.com/links/')

    @mock_links
    def test_get_link(self):
        link = self.browser.get_link()
        assert_equal(link.get('href'), '/link1/')

    @mock_links
    def test_get_links(self):
        links = self.browser.get_links()
        assert_equal(len(links), 3)

    @mock_links
    def test_get_link_by_text(self):
        link = self.browser.get_link('opera')
        assert_equal(link.get('href'), '/link2/')

    @mock_links
    def test_follow_link_tag(self):
        link = self.browser.get_link(text=re.compile('sheer'))
        self.browser.follow_link(link)
        assert_equal(self.browser.url, 'http://robobrowser.com/link1/')

    @mock_links
    def test_follow_link_text(self):
        self.browser.follow_link('heart attack')
        assert_equal(self.browser.url, 'http://robobrowser.com/link1/')

    @mock_links
    def test_follow_link_regex(self):
        self.browser.follow_link(re.compile(r'opera'))
        assert_equal(self.browser.url, 'http://robobrowser.com/link2/')

    @mock_links
    def test_follow_link_bs_args(self):
        self.browser.follow_link(class_=re.compile(r'song'))
        assert_equal(self.browser.url, 'http://robobrowser.com/link2/')

    @mock_links
    def test_follow_link_no_href(self):
        assert_raises(
            exceptions.RoboError,
            lambda: self.browser.follow_link(class_=re.compile(r'nohref'))
        )


class TestForms(unittest.TestCase):

    def setUp(self):
        self.browser = RoboBrowser()

    @mock_forms
    def test_get_forms(self):
        self.browser.open('http://robobrowser.com/get_form/')
        forms = self.browser.get_forms()
        assert_equal(len(forms), 2)

    @mock_forms
    def test_get_form_by_id(self):
        self.browser.open('http://robobrowser.com/get_form/')
        form = self.browser.get_form('bass')
        assert_equal(form.parsed.get('id'), 'bass')

    @mock_forms
    def test_submit_form_get(self):
        self.browser.open('http://robobrowser.com/get_form/')
        form = self.browser.get_form()
        self.browser.submit_form(form)
        assert_equal(
            self.browser.url,
            'http://robobrowser.com/get_form/?deacon=john'
        )
        assert_true(self.browser.state.response.request.body is None)

    @mock_forms
    def test_submit_form_multi_submit(self):
        self.browser.open('http://robobrowser.com/multi_submit_form/')
        form = self.browser.get_form()
        submit = form.submit_fields['submit2']
        self.browser.submit_form(form, submit=submit)
        assert_equal(
            self.browser.url,
            'http://robobrowser.com/multi_submit_form/'
            '?deacon=john&submit2=value2'
        )

    @mock_forms
    def test_submit_form_post(self):
        self.browser.open('http://robobrowser.com/post_form/')
        form = self.browser.get_form()
        self.browser.submit_form(form)
        assert_equal(
            self.browser.url,
            'http://robobrowser.com/submit/'
        )
        assert_equal(
            self.browser.state.response.request.body,
            'deacon=john'
        )


class TestFormsInputNoName(unittest.TestCase):

    @mock_forms
    def setUp(self):
        self.browser = RoboBrowser()
        self.browser.open('http://robobrowser.com/noname/')

    @mock_forms
    def test_get_forms(self):
        forms = self.browser.get_forms()
        assert_equal(len(forms), 1)


class TestHistoryInternals(unittest.TestCase):

    def setUp(self):
        self.browser = RoboBrowser(history=True)

    @mock_urls
    def test_open_appends_to_history(self):
        assert_equal(len(self.browser._states), 0)
        assert_equal(self.browser._cursor, -1)
        self.browser.open('http://robobrowser.com/page1/')
        assert_equal(len(self.browser._states), 1)
        assert_equal(self.browser._cursor, 0)

    @mock_forms
    def test_submit_appends_to_history(self):
        self.browser.open('http://robobrowser.com/get_form/')
        form = self.browser.get_form()
        self.browser.submit_form(form)

        assert_equal(len(self.browser._states), 2)
        assert_equal(self.browser._cursor, 1)

    @mock_urls
    def test_open_clears_history_after_back(self):
        self.browser.open('http://robobrowser.com/page1/')
        self.browser.open('http://robobrowser.com/page2/')
        self.browser.back()
        self.browser.open('http://robobrowser.com/page3/')
        assert_equal(len(self.browser._states), 2)
        assert_equal(self.browser._cursor, 1)

    @mock_urls
    def test_state_deque_max_length(self):
        browser = RoboBrowser(history=5)
        for _ in range(5):
            browser.open('http://robobrowser.com/page1/')
        assert_equal(len(browser._states), 5)
        browser.open('http://robobrowser.com/page2/')
        assert_equal(len(browser._states), 5)

    @mock_urls
    def test_state_deque_no_history(self):
        browser = RoboBrowser(history=False)
        for _ in range(5):
            browser.open('http://robobrowser.com/page1/')
            assert_equal(len(browser._states), 1)
            assert_equal(browser._cursor, 0)


class TestHistory(unittest.TestCase):

    @mock_urls
    def setUp(self):
        self.browser = RoboBrowser(history=True)
        self.browser.open('http://robobrowser.com/page1/')
        self.browser.open('http://robobrowser.com/page2/')
        self.browser.open('http://robobrowser.com/page3/')

    def test_back(self):
        self.browser.back()
        assert_equal(
            self.browser.url,
            'http://robobrowser.com/page2/'
        )

    def test_back_n(self):
        self.browser.back(n=2)
        assert_equal(
            self.browser.url,
            'http://robobrowser.com/page1/'
        )

    def test_forward(self):
        self.browser.back()
        self.browser.forward()
        assert_equal(
            self.browser.url,
            'http://robobrowser.com/page3/'
        )

    def test_forward_n(self):
        self.browser.back(n=2)
        self.browser.forward(n=2)
        assert_equal(
            self.browser.url,
            'http://robobrowser.com/page3/'
        )

    @mock_urls
    def test_open_clears_forward(self):
        self.browser.back(n=2)
        self.browser.open('http://robobrowser.com/page4/')
        assert_equal(
            self.browser._cursor,
            len(self.browser._states) - 1
        )
        assert_raises(
            exceptions.RoboError,
            self.browser.forward
        )

    def test_back_error(self):
        assert_raises(
            exceptions.RoboError,
            self.browser.back,
            5
        )


class TestTimeout(unittest.TestCase):

    @mock.patch('requests.Session.get')
    def test_no_timeout(self, mock_get):
        browser = RoboBrowser()
        browser.open('http://robobrowser.com/')
        mock_get.assert_called_once_with(
            'http://robobrowser.com/', timeout=None, verify=True
        )

    @mock.patch('requests.Session.get')
    def test_instance_timeout(self, mock_get):
        browser = RoboBrowser(timeout=5)
        browser.open('http://robobrowser.com/')
        mock_get.assert_called_once_with(
            'http://robobrowser.com/', timeout=5, verify=True
        )

    @mock.patch('requests.Session.get')
    def test_call_timeout(self, mock_get):
        browser = RoboBrowser(timeout=5)
        browser.open('http://robobrowser.com/', timeout=10)
        mock_get.assert_called_once_with(
            'http://robobrowser.com/', timeout=10, verify=True
        )


class TestVerify(unittest.TestCase):

    @mock.patch('requests.Session.get')
    def test_no_verify(self, mock_get):
        browser = RoboBrowser()
        browser.open('http://robobrowser.com/')
        mock_get.assert_called_once_with(
            'http://robobrowser.com/', verify=True, timeout=None
        )

    @mock.patch('requests.Session.get')
    def test_instance_verify(self, mock_get):
        browser = RoboBrowser(verify=True)
        browser.open('http://robobrowser.com/')
        mock_get.assert_called_once_with(
            'http://robobrowser.com/', verify=True, timeout=None
        )

    @mock.patch('requests.Session.get')
    def test_call_verify(self, mock_get):
        browser = RoboBrowser(verify=True)
        browser.open('http://robobrowser.com/', verify=False)
        mock_get.assert_called_once_with(
            'http://robobrowser.com/', verify=False, timeout=None
        )



"""
Robotic browser.
"""

import re
import requests
from bs4 import BeautifulSoup
from bs4.element import Tag
from requests.exceptions import RequestException

from robobrowser import helpers
from robobrowser import exceptions
from robobrowser.compat import urlparse, string_types
from robobrowser.forms.form import Form
from robobrowser.cache import RoboHTTPAdapter
from robobrowser.helpers import retry


_link_ptn = re.compile(r'^(a|button)$', re.I)
_form_ptn = re.compile(r'^form$', re.I)


class RoboState(object):
    """Representation of a browser state. Wraps the browser and response, and
    lazily parses the response content.

    """

    def __init__(self, browser, response):
        self.browser = browser
        self.response = response
        self.url = response.url
        self._parsed = None

    @property
    def parsed(self):
        """Lazily parse response content, using HTML parser specified by the
        browser.

        """
        if self._parsed is None:
            self._parsed = BeautifulSoup(
                self.response.content,
                features=self.browser.parser,
            )
        return self._parsed


class RoboBrowser(object):
    """Robotic web browser. Represents HTTP requests and responses using the
    requests library and parsed HTML using BeautifulSoup.

    :param tuple auth: Tuple of (username, password)
    :param str parser: HTML parser; used by BeautifulSoup
    :param dict headers: Default headers
    :param str user_agent: Default user-agent
    :param history: History length; infinite if True, 1 if falsy, else
        takes integer value
    :param int timeout: Default timeout in seconds
    :param bool verify: Verify SSL

    :param bool cache: Cache responses
    :param list cache_patterns: List of URL patterns for cache
    :param timedelta max_age: Max age for cache
    :param int max_count: Max count for cache

    :param int tries: Number of retries
    :param Exception errors: Exception(s) to catch
    :param int delay: Delay between retries
    :param int multiplier: Delay multiplier between retries

    """
    def __init__(self, auth=None, parser=None, headers=None, user_agent=None,
                 history=True, timeout=None, verify=True, cache=False,
                 cache_patterns=None, max_age=None, max_count=None, tries=None,
                 errors=RequestException, delay=None, multiplier=None):

        self.session = requests.Session()

        # Add default basic auth
        if auth:
            self.session.auth = auth

        # Add default headers
        headers = headers or {}
        if user_agent is not None:
            headers['User-Agent'] = user_agent
        self.session.headers.update(headers)

        self.parser = parser
        self.timeout = timeout
        self.verify = verify

        # Set up caching
        if cache:
            adapter = RoboHTTPAdapter(max_age=max_age, max_count=max_count)
            cache_patterns = cache_patterns or ['http://', 'https://']
            for pattern in cache_patterns:
                self.session.mount(pattern, adapter)
        elif max_age:
            raise ValueError('Parameter `max_age` is provided, '
                             'but caching is turned off')
        elif max_count:
            raise ValueError('Parameter `max_count` is provided, '
                             'but caching is turned off')

        # Configure history
        self.history = history
        if history is True:
            self._maxlen = None
        elif not history:
            self._maxlen = 1
        else:
            self._maxlen = history
        self._states = []
        self._cursor = -1

        # Set up retries
        if tries:
            decorator = retry(tries, errors, delay, multiplier)
            self._open, self.open = self.open, decorator(self.open)
            self._submit_form, self.submit_form = \
                self.submit_form, decorator(self.submit_form)

    def __repr__(self):
        try:
            return '<RoboBrowser url={0}>'.format(self.url)
        except exceptions.RoboError:
            return '<RoboBrowser>'

    @property
    def state(self):
        if self._cursor == -1:
            raise exceptions.RoboError('No state')
        try:
            return self._states[self._cursor]
        except IndexError:
            raise exceptions.RoboError('Index out of range')

    @property
    def response(self):
        return self.state.response

    @property
    def url(self):
        return self.state.url

    @property
    def parsed(self):
        return self.state.parsed

    @property
    def find(self):
        """See ``BeautifulSoup::find``."""
        try:
            return self.parsed.find
        except AttributeError:
            raise exceptions.RoboError

    @property
    def find_all(self):
        """See ``BeautifulSoup::find_all``."""
        try:
            return self.parsed.find_all
        except AttributeError:
            raise exceptions.RoboError

    @property
    def select(self):
        """See ``BeautifulSoup::select``."""
        try:
            return self.parsed.select
        except AttributeError:
            raise exceptions.RoboError

    def _build_url(self, url):
        """Build absolute URL.

        :param url: Full or partial URL
        :return: Full URL

        """
        return urlparse.urljoin(
            self.url,
            url
        )

    def open(self, url, timeout=None, verify=None):
        """Open a URL.

        :param str url: URL
        :param int timeout: Timeout in seconds
        :param bool verify: Verify SSL

        """
        response = self.session.get(
            url,
            timeout=timeout or self.timeout,
            verify=verify if verify is not None else self.verify,
        )
        self._update_state(response)

    def _update_state(self, response):
        """Update the state of the browser. Create a new state object, and
        append to or overwrite the browser's state history.

        :param requests.MockResponse: New response object

        """
        # Clear trailing states
        self._states = self._states[:self._cursor + 1]

        # Append new state
        state = RoboState(self, response)
        self._states.append(state)
        self._cursor += 1

        # Clear leading states
        if self._maxlen:
            decrement = len(self._states) - self._maxlen
            if decrement > 0:
                self._states = self._states[decrement:]
                self._cursor -= decrement

    def _traverse(self, n=1):
        """Traverse state history. Used by `back` and `forward` methods.

        :param int n: Cursor increment. Positive values move forward in the
            browser history; negative values move backward.

        """
        if not self.history:
            raise exceptions.RoboError('Not tracking history')
        cursor = self._cursor + n
        if cursor >= len(self._states) or cursor < 0:
            raise exceptions.RoboError('Index out of range')
        self._cursor = cursor

    def back(self, n=1):
        """Go back in browser history.

        :param int n: Number of pages to go back

        """
        self._traverse(-1 * n)

    def forward(self, n=1):
        """Go forward in browser history.

        :param int n: Number of pages to go forward

        """
        self._traverse(n)

    def get_link(self, text=None, *args, **kwargs):
        """Find an anchor or button by containing text, as well as standard
        BeautifulSoup arguments.

        :param text: String or regex to be matched in link text
        :return: BeautifulSoup tag if found, else None

        """
        return helpers.find(
            self.parsed, _link_ptn, text=text, *args, **kwargs
        )

    def get_links(self, text=None, *args, **kwargs):
        """Find anchors or buttons by containing text, as well as standard
        BeautifulSoup arguments.

        :param text: String or regex to be matched in link text
        :return: List of BeautifulSoup tags

        """
        return helpers.find_all(
            self.parsed, _link_ptn, text=text, *args, **kwargs
        )

    def get_form(self, id=None, *args, **kwargs):
        """Find form by ID, as well as standard BeautifulSoup arguments.

        :param str id: Form ID
        :return: BeautifulSoup tag if found, else None

        """
        if id:
            kwargs['id'] = id
        form = self.find(_form_ptn, *args, **kwargs)
        if form is not None:
            return Form(form)

    def get_forms(self, *args, **kwargs):
        """Find forms by standard BeautifulSoup arguments.

        :return: List of BeautifulSoup tags

        """
        forms = self.find_all(_form_ptn, *args, **kwargs)
        return [
            Form(form)
            for form in forms
        ]

    def follow_link(self, value=None, *args, **kwargs):
        """Find a click a link by tag, pattern, and/or BeautifulSoup
        arguments.

        :param value: BeautifulSoup tag, string, or regex. If tag, follow its
            href; if string or regex, search parsed document for match.

        """
        if isinstance(value, Tag):
            link = value
        elif isinstance(value, string_types):
            link = self.get_link(text=value, *args, **kwargs)
        elif isinstance(value, re._pattern_type):
            link = self.get_link(text=value, *args, **kwargs)
        else:
            link = self.get_link(*args, **kwargs)
        if link is None:
            raise exceptions.RoboError('No results found')
        href = link.get('href')
        if href is None:
            raise exceptions.RoboError('Link element must have href attribute')
        self.open(self._build_url(href))

    def submit_form(self, form, submit=None):
        """Submit a form.

        :param Form form: Filled-out form object
        :param Submit submit: Optional `Submit` to click, if form includes
            multiple submits

        """
        # Get HTTP verb
        method = form.method.upper()

        # Send request
        url = self._build_url(form.action) or self.url
        payload = form.serialize(submit=submit)
        response = self.session.request(method, url, **payload.to_requests(method))

        # Update history
        self._update_state(response)



import mock
import unittest
from nose.tools import *

import tempfile
from bs4 import BeautifulSoup

from robobrowser.compat import builtin_name
from robobrowser.forms.form import Form, Payload, fields, _parse_fields
from robobrowser import exceptions


class TestPayload(unittest.TestCase):

    def setUp(self):
        self.payload = Payload()
        self.payload.add({'red': 'special'})

    def test_add_payload(self):
        self.payload.add({'lazing': 'sunday'})
        assert_true('lazing' in self.payload.data)
        assert_equal(self.payload.data['lazing'], 'sunday')

    def test_add_by_key(self):
        self.payload.add({'lazing': 'sunday'}, 'afternoon')
        assert_false('lazing' in self.payload.data)
        assert_true('afternoon' in self.payload.options)
        assert_true('lazing' in self.payload.options['afternoon'])
        assert_equal(
            self.payload.options['afternoon']['lazing'],
            'sunday'
        )

    def test_requests_get(self):
        out = self.payload.to_requests('get')
        assert_true('params' in out)
        assert_equal(list(out['params']), [('red', 'special')])

    def test_requests_post(self):
        out = self.payload.to_requests('post')
        assert_true('data' in out)
        assert_equal(list(out['data']), [('red', 'special')])


class TestForm(unittest.TestCase):

    def setUp(self):
        self.html = '''
            <form>
                <input name="vocals" />
                <input name="guitar" type="file" />
                <select name="drums">
                    <option value="roger">Roger<br />
                    <option value="john">John<br />
                </select>
                <input type="radio" name="bass" value="Roger">Roger<br />
                <input type="radio" name="bass" value="John">John<br />
                <input name="multi" value="multi1" />
                <input name="multi" value="multi2" />
                <input type="submit" name="submit" value="submit" />
            </form>
        '''
        self.form = Form(self.html)

    def test_fields(self):
        keys = set(('vocals', 'guitar', 'drums', 'bass', 'multi', 'submit'))
        assert_equal(set(self.form.fields.keys()), keys)
        assert_equal(set(self.form.keys()), keys)

    def test_add_field(self):
        html = '<input name="instrument" />'
        field = fields.Input(html)
        self.form.add_field(field)
        assert_true('instrument' in self.form.fields)

    def test_add_field_wrong_type(self):
        assert_raises(ValueError, lambda: self.form.add_field('freddie'))

    def test_repr(self):
        assert_equal(
            repr(self.form),
            '<RoboForm vocals=, guitar=, drums=roger, bass=, '
                'multi=multi1, multi=multi2, submit=submit>'
        )

    def test_repr_empty(self):
        assert_equal(
            repr(Form('<form></form>')),
            '<RoboForm>'
        )

    def test_serialize(self):
        serialized = self.form.serialize()
        assert_equal(serialized.data.getlist('multi'), ['multi1', 'multi2'])
        assert_equal(serialized.data['submit'], 'submit')


class TestFormMultiSubmit(unittest.TestCase):

    def setUp(self):
        self.html = '''
            <form>
                <input type="submit" name="submit1" value="value1" />
                <input type="submit" name="submit2" value="value2" />
            </form>
        '''
        self.form = Form(self.html)

    def test_serialize_multi_no_submit_specified(self):
        assert_raises(
            exceptions.InvalidSubmitError,
            lambda: self.form.serialize()
        )

    def test_serialize_multi_wrong_submit_specified(self):
        fake_submit = fields.Submit('<input type="submit" name="fake" />')
        assert_raises(
            exceptions.InvalidSubmitError,
            lambda: self.form.serialize(submit=fake_submit)
        )

    def test_serialize_multi(self):
        submit = self.form.submit_fields['submit1']
        serialized = self.form.serialize(submit)
        assert_equal(serialized.data['submit1'], 'value1')
        assert_false('submit2' in serialized.data)


class TestParser(unittest.TestCase):

    def setUp(self):
        self.form = Form('<form></form>')

    def test_method_default(self):
        assert_equal(self.form.method, 'get')

    def test_method(self):
        form = Form('<form method="put"></form>')
        assert_equal(form.method, 'put')

    def test_action(self):
        form = Form('<form action="/"></form>')
        assert_equal(form.action, '/')

    def test_parse_input(self):
        html = '<input name="band" value="queen" />'
        _fields = _parse_fields(BeautifulSoup(html))
        assert_equal(len(_fields), 1)
        assert_true(isinstance(_fields[0], fields.Input))

    def test_parse_file_input(self):
        html = '<input name="band" type="file" />'
        _fields = _parse_fields(BeautifulSoup(html))
        assert_equal(len(_fields), 1)
        assert_true(isinstance(_fields[0], fields.FileInput))

    def test_parse_textarea(self):
        html = '<textarea name="band">queen</textarea>'
        _fields = _parse_fields(BeautifulSoup(html))
        assert_equal(len(_fields), 1)
        assert_true(isinstance(_fields[0], fields.Textarea))

    def test_parse_radio(self):
        html = '''
            <input type="radio" name="favorite_member" />freddie<br />
            <input type="radio" name="favorite_member" />brian<br />
            <input type="radio" name="favorite_member" />roger<br />
            <input type="radio" name="favorite_member" />john<br />
            <input type="radio" name="favorite_song" />rhapsody<br />
            <input type="radio" name="favorite_song" />killer<br />
        '''
        _fields = _parse_fields(BeautifulSoup(html))
        assert_equal(len(_fields), 2)
        assert_true(isinstance(_fields[0], fields.Radio))
        assert_true(isinstance(_fields[0], fields.Radio))
        assert_equal(
            len(_fields[0]._parsed), 4
        )
        assert_equal(
            len(_fields[1]._parsed), 2
        )

    def test_parse_checkbox(self):
        html = '''
            <input type="checkbox" name="favorite_member" />freddie<br />
            <input type="checkbox" name="favorite_member" />brian<br />
            <input type="checkbox" name="favorite_member" />roger<br />
            <input type="checkbox" name="favorite_member" />john<br />
            <input type="checkbox" name="favorite_song" />rhapsody<br />
            <input type="checkbox" name="favorite_song" />killer<br />
        '''
        _fields = _parse_fields(BeautifulSoup(html))
        assert_equal(len(_fields), 2)
        assert_true(isinstance(_fields[0], fields.Checkbox))
        assert_true(isinstance(_fields[1], fields.Checkbox))
        assert_equal(len(_fields[0]._parsed), 4)
        assert_equal(len(_fields[1]._parsed), 2)

    def test_parse_select(self):
        html = '''
            <select name="instrument">
                <option value="vocals">vocals</option>
                <option value="guitar">guitar</option>
                <option value="drums">drums</option>
                <option value="bass">bass</option>
            </select>
        '''
        _fields = _parse_fields(BeautifulSoup(html))
        assert_equal(len(_fields), 1)
        assert_true(isinstance(_fields[0], fields.Select))

    def test_parse_select_multi(self):
        html = '''
            <select name="instrument" multiple>
                <option value="vocals">vocals</option>
                <option value="guitar">guitar</option>
                <option value="drums">drums</option>
                <option value="bass">bass</option>
            </select>
        '''
        _fields = _parse_fields(BeautifulSoup(html))
        assert_equal(len(_fields), 1)
        assert_true(isinstance(_fields[0], fields.MultiSelect))


class TestInput(unittest.TestCase):

    def setUp(self):
        self.html = '<input name="brian" value="may" />'
        self.input = fields.Input(BeautifulSoup(self.html).find('input'))

    def test_name(self):
        assert_equal(self.input.name, 'brian')

    def test_initial(self):
        assert_equal(self.input._value, 'may')
        assert_equal(self.input.value, 'may')

    def test_value(self):
        self.input.value = 'red special'
        assert_equal(self.input._value, 'red special')
        assert_equal(self.input.value, 'red special')

    def test_serialize(self):
        assert_equal(
            self.input.serialize(),
            {'brian': 'may'}
        )

    def test_invalid_name(self):
        html = '<input />'
        assert_raises(exceptions.InvalidNameError, lambda: fields.Input(html))


class TestInputBlank(unittest.TestCase):

    def setUp(self):
        self.html = '<input name="blank" />'
        self.input = fields.Input(BeautifulSoup(self.html).find('input'))

    def test_initial(self):
        assert_equal(self.input._value, None)
        assert_equal(self.input.value, '')

    def test_serialize(self):
        assert_equal(
            self.input.serialize(),
            {'blank': ''}
        )


class TestTextarea(unittest.TestCase):

    def setUp(self):
        self.html = '<textarea name="roger">taylor</textarea>'
        self.input = fields.Textarea(BeautifulSoup(self.html).find('textarea'))

    def test_name(self):
        assert_equal(self.input.name, 'roger')

    def test_initial(self):
        assert_equal(self.input._value, 'taylor')
        assert_equal(self.input.value, 'taylor')

    def test_value(self):
        self.input.value = 'the drums'
        assert_equal(self.input._value, 'the drums')
        assert_equal(self.input.value, 'the drums')

    def test_serialize(self):
        assert_equal(
            self.input.serialize(),
            {'roger': 'taylor'}
        )


class TestTextareaBlank(unittest.TestCase):

    def setUp(self):
        self.html = '<textarea name="blank"></textarea>'
        self.input = fields.Textarea(BeautifulSoup(self.html).find('textarea'))

    def test_initial(self):
        assert_equal(self.input._value, '')
        assert_equal(self.input.value, '')

    def test_serialize(self):
        assert_equal(
            self.input.serialize(),
            {'blank': ''}
        )


class TestSelect(unittest.TestCase):

    def setUp(self):
        self.html = '''
            <select name="john">
                <option value="tie">your mother down</option>
                <option value="you're" selected>my best friend</option>
                <option value="the">millionaire waltz</option>
            </select>
        '''
        self.input = fields.Select(BeautifulSoup(self.html).find('select'))

    def test_name(self):
        assert_equal(self.input.name, 'john')

    def test_options(self):
        assert_equal(
            self.input.options,
            ['tie', "you're", 'the']
        )

    def test_initial(self):
        assert_equal(self.input._value, 1)
        assert_equal(self.input.value, "you're")

    def test_value(self):
        self.input.value = 'the'
        assert_equal(self.input._value, 2)
        assert_equal(self.input.value, 'the')

    def test_value_label(self):
        self.input.value = 'millionaire waltz'
        assert_equal(self.input._value, 2)
        assert_equal(self.input.value, 'the')

    def test_serialize(self):
        assert_equal(
            self.input.serialize(),
            {'john': "you're"}
        )


class TestSelectBlank(unittest.TestCase):

    def setUp(self):
        self.html = '''
            <select name="john">
                <option value="tie">your mother down</option>
                <option value="you're">my best friend</option>
                <option value="the">millionaire waltz</option>
            </select>
        '''
        self.input = fields.Select(BeautifulSoup(self.html).find('select'))

    def test_name(self):
        assert_equal(self.input.name, 'john')

    def test_initial(self):
        assert_equal(self.input._value, 0)
        assert_equal(self.input.value, 'tie')

    def test_serialize(self):
        assert_equal(
            self.input.serialize(),
            {'john': 'tie'}
        )


class TestMultiSelect(unittest.TestCase):

    def setUp(self):
        self.html = '''
            <select name="john" multiple>
                <option value="tie">your mother down</option>
                <option value="you're" selected>my best friend</option>
                <option value="the">millionaire waltz</option>
            </select>
        '''
        self.input = fields.MultiSelect(BeautifulSoup(self.html).find('select'))


class TestMixedCase(unittest.TestCase):

    def test_upper_type(self):
        html = '''
            <input type="RADIO" name="members" value="mercury" />vocals<br />
        '''
        input = fields.Radio(BeautifulSoup(html).find_all('input'))
        assert_equal(input.name, 'members')

    def test_upper_name(self):
        html = '''
            <input type="radio" NAME="members" value="mercury" />vocals<br />
        '''
        input = fields.Radio(BeautifulSoup(html).find_all('input'))
        assert_equal(input.name, 'members')

    def test_mixed_radio_names(self):
        html = '''
            <input type="radio" NAME="members" value="mercury" />vocals<br />
            <input type="radio" NAME="MEMBERS" value="may" />guitar<br />
        '''
        input = fields.Radio(BeautifulSoup(html).find_all('input'))
        assert_equal(input.name, 'members')
        assert_equal(
            input.options,
            ['mercury', 'may']
        )


class TestRadio(unittest.TestCase):

    def setUp(self):
        self.html = '''
            <input type="radio" name="members" value="mercury" checked />vocals<br />
            <input type="radio" name="members" value="may" />guitar<br />
            <input type="radio" name="members" value="taylor" />drums<br />
            <input type="radio" name="members" value="deacon" checked />bass<br />
        '''
        self.input = fields.Radio(BeautifulSoup(self.html).find_all('input'))

    def test_name(self):
        assert_equal(self.input.name, 'members')

    def test_options(self):
        assert_equal(
            self.input.options,
            ['mercury', 'may', 'taylor', 'deacon']
        )

    def test_initial(self):
        assert_equal(self.input.value, 'mercury')

    def test_value(self):
        self.input.value = 'taylor'
        assert_equal(self.input._value, 2)
        assert_equal(self.input.value, 'taylor')

    def test_value_label(self):
        self.input.value = 'drums'
        assert_equal(self.input._value, 2)
        assert_equal(self.input.value, 'taylor')

    def test_serialize(self):
        assert_equal(
            self.input.serialize(),
            {'members': 'mercury'}
        )


class TestRadioBlank(unittest.TestCase):

    def setUp(self):
        self.html = '''
            <input type="radio" name="member" value="mercury" />vocals<br />
            <input type="radio" name="member" value="may" />guitar<br />
            <input type="radio" name="member" value="taylor" />drums<br />
            <input type="radio" name="member" value="deacon" />bass<br />
        '''
        self.input = fields.Radio(BeautifulSoup(self.html).find_all('input'))

    def test_initial(self):
        assert_equal(self.input.value, '')

    def test_serialize(self):
        assert_equal(
            self.input.serialize(),
            {'member': ''}
        )


class TestCheckbox(unittest.TestCase):

    def setUp(self):
        self.html = '''
            <input type="checkbox" name="member" value="mercury" checked />vocals<br />
            <input type="checkbox" name="member" value="may" />guitar<br />
            <input type="checkbox" name="member" value="taylor" />drums<br />
            <input type="checkbox" name="member" value="deacon" checked />bass<br />
        '''
        self.input = fields.Checkbox(BeautifulSoup(self.html).find_all('input'))

    def test_name(self):
        assert_equal(self.input.name, 'member')

    def test_options(self):
        assert_equal(
            self.input.options,
            ['mercury', 'may', 'taylor', 'deacon']
        )

    def test_initial(self):
        assert_equal(
            self.input.value,
            ['mercury', 'deacon']
        )

    def test_value(self):
        self.input.value = 'taylor'
        assert_equal(self.input._value, [2])
        assert_equal(self.input.value, ['taylor'])

    def test_value(self):
        self.input.value = ['taylor', 'deacon']
        assert_equal(self.input._value, [2, 3])
        assert_equal(self.input.value, ['taylor', 'deacon'])

    def test_value_label(self):
        self.input.value = 'drums'
        assert_equal(self.input._value, [2])
        assert_equal(self.input.value, ['taylor'])

    def test_serialize(self):
        assert_equal(
            self.input.serialize(),
            {'member': ['mercury', 'deacon']}
        )


class TestCheckboxBlank(unittest.TestCase):

    def setUp(self):
        self.html = '''
            <input type="checkbox" name="member" value="mercury" />vocals<br />
            <input type="checkbox" name="member" value="may" />guitar<br />
            <input type="checkbox" name="member" value="taylor" />drums<br />
            <input type="checkbox" name="member" value="deacon" />bass<br />
        '''
        self.input = fields.Checkbox(BeautifulSoup(self.html).find_all('input'))

    def test_initial(self):
        assert_equal(
            self.input.value, []
        )

    def test_serialize(self):
        assert_equal(
            self.input.serialize(),
            {'member': []}
        )


class TestFileInput(unittest.TestCase):

    def setUp(self):
        self.html = '<input name="song" type="file" />'
        self.input = fields.FileInput(BeautifulSoup(self.html).find('input'))

    def test_name(self):
        assert_equal(self.input.name, 'song')

    def test_value_file(self):
        file = tempfile.TemporaryFile('r')
        self.input.value = file
        assert_equal(self.input._value, file)
        assert_equal(self.input.value, file)

    @mock.patch('{0}.open'.format(builtin_name))
    def test_value_name(self, mock_open):
        file = tempfile.TemporaryFile('r')
        mock_open.return_value = file
        self.input.value = 'temp'
        assert_equal(self.input._value, file)
        assert_equal(self.input.value, file)

    def test_serialize(self):
        file = tempfile.TemporaryFile('r')
        self.input.value = file
        assert_equal(
            self.input.serialize(),
            {'song': file}
        )


class TestDefaultValues(unittest.TestCase):

    def test_checkbox_default(self):
        inputs = BeautifulSoup('''
            <input type="checkbox" name="checkbox" />
        ''').find_all('input')
        checkbox = fields.Checkbox(inputs)
        assert_equal(checkbox.options, ['on'])

    def test_radio_default(self):
        inputs = BeautifulSoup('''
            <input type="radio" name="checkbox" />
        ''').find_all('input')
        radio = fields.Radio(inputs)
        assert_equal(radio.options, ['on'])

    def test_select_default(self):
        parsed = BeautifulSoup('''
            <select name="select">
                <option>opt</option>
            </select>
        ''')
        select = fields.Select(parsed)
        assert_equal(select.options, ['sel'])

    def test_multi_select_default(self):
        parsed = BeautifulSoup('''
            <select name="select" multiple>
                <option>opt</option>
            </select>
        ''')
        select = fields.Select(parsed)
        assert_equal(select.options, ['sel'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re
import sys

from setuptools import setup, find_packages
from pip.req import parse_requirements

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

def find_version(fname):
    """Attempts to find the version number in the file names fname.
    Raises RuntimeError if not found.

    """
    version = ''
    with open(fname, 'r') as fp:
        reg = re.compile(r'__version__ = [\'"]([^\'"]*)[\'"]')
        for line in fp:
            m = reg.match(line)
            if m:
                version = m.group(1)
                break
    if not version:
        raise RuntimeError('Cannot find version information')
    return version

__version__ = find_version('robobrowser/__init__.py')

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')
requirements = [
    str(requirement.req)
    for requirement in parse_requirements('requirements.txt')
]

setup(
    name='robobrowser',
    version=__version__,
    description='Your friendly neighborhood web scraper',
    author='Joshua Carp',
    author_email='jm.carp@gmail.com',
    url='https://github.com/jmcarp/robobrowser',
    packages=find_packages(),
    package_dir={'robobrowser': 'robobrowser'},
    include_package_data=True,
    install_requires=requirements,
    tests_require=[
        'nose',
    ],
    license='MIT',
    zip_safe=False,
    keywords='robobrowser',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    test_suite='tests',
)
